from django import template
from django.template.defaultfilters import stringfilter

from machina.models.fields import render_func


register = template.Library()


@register.filter(is_safe=True)
@stringfilter
def rendered(value):
    return render_func(value)



"""
    Forum forms
    ===========

    This module defines forms provided by the ``forum`` application.

"""

from django import forms
from django.contrib.admin.widgets import ForeignKeyRawIdWidget
from django.forms.forms import NON_FIELD_ERRORS
from django.utils.translation import ugettext_lazy as _
from mptt.forms import TreeNodeChoiceField

from machina.core.db.models import get_model
from machina.core.loading import get_class


Forum = get_model('forum', 'Forum')
GroupForumPermission = get_model('forum_permission', 'GroupForumPermission')
UserForumPermission = get_model('forum_permission', 'UserForumPermission')

PermissionConfig = get_class('forum_permission.defaults', 'PermissionConfig')


class PickUserForm(forms.Form):
    """ Form allowing to pick a user to edit their permissions. """

    user = UserForumPermission._meta.get_field('user').formfield()
    anonymous_user = forms.BooleanField(
        label=_('Anonymous'),
        initial=False,
        help_text=_(
            'Please select this option if you want to edit the permissions of the anonymous user'
        ),
    )

    def __init__(self, *args, **kwargs):
        admin_site = kwargs.pop('admin_site')
        super().__init__(*args, **kwargs)

        self.fields['user'].required = False
        self.fields['user'].widget = ForeignKeyRawIdWidget(
            UserForumPermission._meta.get_field('user').remote_field, admin_site,
        )

        self.fields['anonymous_user'].required = False

    def clean(self):
        cleaned_data = super().clean()
        user = cleaned_data.get('user', None)
        anonymous_user = cleaned_data.get('anonymous_user', None)
        if user and anonymous_user:
            self._errors[NON_FIELD_ERRORS] = self.error_class(
                [_('Choose either a user ID or select the anonymous user'), ],
            )
        return cleaned_data


class PickGroupForm(forms.Form):
    """ Form allowing to pick a group to edit its permissions. """

    group = GroupForumPermission._meta.get_field('group').formfield()

    def __init__(self, *args, **kwargs):
        admin_site = kwargs.pop('admin_site')
        super().__init__(*args, **kwargs)

        self.fields['group'].required = False
        self.fields['group'].widget = ForeignKeyRawIdWidget(
            GroupForumPermission._meta.get_field('group').remote_field, admin_site,
        )


class PickForumForm(forms.Form):
    """ Form allowing to pick a specific forum. """

    forum = TreeNodeChoiceField(queryset=Forum.objects.all(), required=False)


class PermissionsForm(forms.Form):
    """ Form allowing to edit permissions. """

    PERM_GRANTED = 'granted'
    PERM_NOT_GRANTED = 'not-granted'
    PERM_NOT_SET = 'not-set'

    def __init__(self, *args, **kwargs):
        self.permissions_dict = kwargs.pop('permissions_dict', {})
        super().__init__(*args, **kwargs)

        # Initializes permission fields
        f_choices = (
            (self.PERM_NOT_SET, _('Not set')),
            (self.PERM_GRANTED, _('Granted')),
            (self.PERM_NOT_GRANTED, _('Not granted')),
        )
        for scope in PermissionConfig.scopes:
            codenames = [
                x['codename'] for x in PermissionConfig.permissions if x['scope'] == scope
            ]
            permissions = filter(lambda v: v[0] in codenames, self.permissions_dict.items())
            for codename, p in permissions:
                self.fields[codename] = forms.ChoiceField(
                    label=p[0].name, choices=f_choices, required=False, widget=forms.RadioSelect,
                )
                self.fields[codename].initial = p[1]
                self.fields[codename].scope = scope



# -*- coding: utf-8 -*-

from .attachments import *  # noqa
from .auth import *  # noqa
from .conversation import *  # noqa
from .forum import *  # noqa
from .permission import *  # noqa
from .polls import *  # noqa
from .tracking import *  # noqa



from django import template

from machina.core.loading import get_class


get_forum_member_display_name = get_class('forum_member.shortcuts', 'get_forum_member_display_name')

register = template.Library()


@register.filter
def forum_member_display_name(user):
    """ Returns the forum member display name to use for a given user.

    Usage::

        {{ user|forum_member_display_name }}

    """
    return get_forum_member_display_name(user)



# Generated by Django 2.1.2 on 2018-11-03 19:04

from django.db import migrations, models
import machina.apps.forum_conversation.forum_attachments.abstract_models


class Migration(migrations.Migration):

    dependencies = [
        ('forum_attachments', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attachment',
            name='file',
            field=models.FileField(upload_to=machina.apps.forum_conversation.forum_attachments.abstract_models.get_attachment_file_upload_to, verbose_name='File'),
        ),
    ]



"""
    Forum polls views
    =================

    This module defines views provided by the ``forum_polls`` application.

"""

from django.contrib import messages
from django.forms.forms import NON_FIELD_ERRORS
from django.http import HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from django.views.generic import UpdateView
from django.views.generic.edit import ModelFormMixin

from machina.core.db.models import get_model
from machina.core.loading import get_class


TopicPoll = get_model('forum_polls', 'TopicPoll')
TopicPollVote = get_model('forum_polls', 'TopicPollVote')

TopicPollVoteForm = get_class('forum_polls.forms', 'TopicPollVoteForm')

PermissionRequiredMixin = get_class('forum_permission.viewmixins', 'PermissionRequiredMixin')


class TopicPollVoteView(PermissionRequiredMixin, UpdateView):
    """ Allows to vote in polls. """

    form_class = TopicPollVoteForm
    http_method_names = ['post', ]
    model = TopicPoll

    def get_form_kwargs(self):
        """ Returns the keyword arguments to provide tp the associated form. """
        kwargs = super(ModelFormMixin, self).get_form_kwargs()
        kwargs['poll'] = self.object
        return kwargs

    def form_valid(self, form):
        """ Handles a valid form. """
        user_kwargs = (
            {'voter': self.request.user}
            if self.request.user.is_authenticated
            else {'anonymous_key': self.request.user.forum_key}
        )

        if self.object.user_changes:
            # If user changes are allowed for this poll, all the poll associated with the current
            # user must be deleted.
            TopicPollVote.objects.filter(poll_option__poll=self.object, **user_kwargs).delete()

        options = form.cleaned_data['options']
        for option in options:
            TopicPollVote.objects.create(poll_option=option, **user_kwargs)

        return HttpResponseRedirect(self.get_success_url())

    def form_invalid(self, form):
        """ Handles an invalid form. """
        messages.error(self.request, form.errors[NON_FIELD_ERRORS])
        return redirect(
            reverse(
                'forum_conversation:topic',
                kwargs={
                    'forum_slug': self.object.topic.forum.slug,
                    'forum_pk': self.object.topic.forum.pk,
                    'slug': self.object.topic.slug,
                    'pk': self.object.topic.pk
                },
            ),
        )

    def get_success_url(self):
        """ Returns the success URL to redirect the user to. """
        messages.success(self.request, _('Your vote has been cast.'))
        return reverse(
            'forum_conversation:topic',
            kwargs={
                'forum_slug': self.object.topic.forum.slug,
                'forum_pk': self.object.topic.forum.pk,
                'slug': self.object.topic.slug,
                'pk': self.object.topic.pk,
            },
        )

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_object()

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permission check. """
        return self.request.forum_permission_handler.can_vote_in_poll(obj, user)



"""
    Forum conversation signals
    ==========================

    This modules defines Django signals that can be triggered by the ``forum`` application.

"""

import django.dispatch


topic_viewed = django.dispatch.Signal(providing_args=["topic", "user", "request", "response", ])



"""
    Machina core URL helpers
    ========================

    This module defines commons URLs helpers and abstractions used by django-machina to manage its
    URL patterns.

"""


class URLPatternsFactory:
    """ Allows to generate the URL patterns of a machina application.

    Machina's views and URLs use a tree of ``URLPatternsFactory`` instances in order to build the
    complete list of URL patterns of the forum application. This class provides a
    ``get_urlpatterns`` method that allows to define the URL patterns to include in the global list
    of URLs associated with the forum application. This behavior allows to easily override each of
    the views used to build the final list of URL patterns (because the views are imported
    dynamically). It also allows to append additional URLs to the generated URL patterns by
    subclassing the ``URLPatternsFactory`` subclass associated with the considered forum app.

    """

    app_namespace = None

    def get_urlpatterns(self):
        """ Returns the URL patterns managed by the considered factory / application.

        This method should be implemented in any subclass.
        """
        return []

    @property
    def urlpatterns(self):
        """ Returns a tuple containing the list of URL patterns and an app namespace. """
        return self.get_urlpatterns(), self.app_namespace



from django.shortcuts import _get_queryset


def get_object_or_none(klass, *args, **kwargs):
    """ Calls get() on a given model manager, but it returns None instead of the model’s
        DoesNotExist exception.
    """
    queryset = _get_queryset(klass)
    try:
        return queryset.get(*args, **kwargs)
    except:  # noqa: E722
        return None



"""
    Forum conversation URLs
    =======================

    This module defines URL patterns associated with the django-machina's ``forum_conversation``
    application.

"""

from django.urls import include, path
from django.utils.translation import ugettext_lazy as _

from machina.apps.forum_conversation.forum_attachments.urls import \
    urlpatterns_factory as attachments_urlpatterns_factory
from machina.apps.forum_conversation.forum_polls.urls import \
    urlpatterns_factory as polls_urlpatterns_factory
from machina.core.loading import get_class
from machina.core.urls import URLPatternsFactory


class BaseForumConversationURLPatternsFactory(URLPatternsFactory):
    """ Allows to generate the URL patterns of the ``forum_conversation`` application. """

    app_namespace = 'forum_conversation'

    topic_view = get_class('forum_conversation.views', 'TopicView')
    topic_create_view = get_class('forum_conversation.views', 'TopicCreateView')
    topic_update_view = get_class('forum_conversation.views', 'TopicUpdateView')
    post_create_view = get_class('forum_conversation.views', 'PostCreateView')
    post_update_view = get_class('forum_conversation.views', 'PostUpdateView')
    post_delete_view = get_class('forum_conversation.views', 'PostDeleteView')

    def get_urlpatterns(self):
        """ Returns the URL patterns managed by the considered factory / application. """
        urlpatterns = super().get_urlpatterns()

        conversation_urlpatterns = [
            path(
                _('topic/<str:slug>-<int:pk>/'),
                self.topic_view.as_view(),
                name='topic',
            ),
            path(_('topic/create/'), self.topic_create_view.as_view(), name='topic_create'),
            path(
                _('topic/<str:slug>-<int:pk>/update/'),
                self.topic_update_view.as_view(),
                name='topic_update',
            ),
            path(
                _('topic/<str:topic_slug>-<int:topic_pk>/post/create/'),
                self.post_create_view.as_view(),
                name='post_create',
            ),
            path(
                _('topic/<str:topic_slug>-<int:topic_pk>/<int:pk>/post/update/'),
                self.post_update_view.as_view(),
                name='post_update',
            ),
            path(
                _('topic/<str:topic_slug>-<int:topic_pk>/<int:pk>/post/delete/'),
                self.post_delete_view.as_view(),
                name='post_delete',
            ),
        ]

        urlpatterns += [
            path(
                _('forum/<str:forum_slug>-<int:forum_pk>/'),
                include(conversation_urlpatterns),
            ),
        ]

        return urlpatterns


class ForumAttachmentsURLPatternsFactory(URLPatternsFactory):
    """ Allows to generate the URL patterns of the ``forum_attachments`` application. """

    attachments_urlpatterns_factory = attachments_urlpatterns_factory

    def get_urlpatterns(self):
        """ Returns the URL patterns managed by the considered factory / application. """
        return super().get_urlpatterns() + [
            path('', include(self.attachments_urlpatterns_factory.urlpatterns)),
        ]


class ForumPollsURLPatternsFactory(URLPatternsFactory):
    """ Allows to generate the URL patterns of the ``forum_attachments`` application. """

    polls_urlpatterns_factory = polls_urlpatterns_factory

    def get_urlpatterns(self):
        """ Returns the URL patterns managed by the considered factory / application. """
        return super().get_urlpatterns() + [
            path('', include(self.polls_urlpatterns_factory.urlpatterns)),
        ]


class ForumConversationURLPatternsFactory(
    BaseForumConversationURLPatternsFactory, ForumAttachmentsURLPatternsFactory,
    ForumPollsURLPatternsFactory,
):
    """ Composite class combining conversation views with polls & attachments views. """


urlpatterns_factory = ForumConversationURLPatternsFactory()



"""
    Forum permission shortcuts
    ==========================

    This module defines shortcut functions allowing to easily perform permission checks and to
    assign or remove granted permissions.

"""

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group

from machina.core.db.models import get_model


ForumPermission = get_model('forum_permission', 'ForumPermission')
GroupForumPermission = get_model('forum_permission', 'GroupForumPermission')
UserForumPermission = get_model('forum_permission', 'UserForumPermission')


class NotUserNorGroup(Exception):
    pass


def assign_perm(perm, user_or_group, forum=None, has_perm=True):
    """ Assigns a permission to a user (anonymous or not) or a group. """
    user, group = get_identity(user_or_group)
    perm = ForumPermission.objects.get(codename=perm)
    if user:
        return UserForumPermission.objects.create(
            forum=forum,
            permission=perm,
            user=user if not user.is_anonymous else None,
            anonymous_user=user.is_anonymous,
            has_perm=has_perm,
        )
    if group:
        return GroupForumPermission.objects.create(
            forum=forum, permission=perm, group=group, has_perm=has_perm,
        )


def remove_perm(perm, user_or_group, forum=None):
    """ Remove a permission to a user (anonymous or not) or a group. """
    user, group = get_identity(user_or_group)
    perm = ForumPermission.objects.get(codename=perm)
    if user:
        UserForumPermission.objects.filter(
            forum=forum,
            permission=perm,
            user=user if not user.is_anonymous else None,
            anonymous_user=user.is_anonymous,
        ).delete()
    if group:
        GroupForumPermission.objects.filter(forum=forum, permission=perm, group=group).delete()


def get_identity(identity):
    """ Returns a (user_obj, None) tuple or a (None, group_obj) tuple depending on the considered
        instance.
    """
    if isinstance(identity, AnonymousUser):
        return identity, None

    if isinstance(identity, get_user_model()):
        return identity, None
    elif isinstance(identity, Group):
        return None, identity
    else:  # pragma: no cover
        raise NotUserNorGroup(
            'User/AnonymousUser or Group instance is required '
            '(got {})'.format(identity),
        )


def get_anonymous_user_forum_key(user):
    """ Returns the forum key identifier associated with the considered anonymous user. """
    return (
        user.forum_key if isinstance(user, AnonymousUser) and hasattr(user, 'forum_key')
        else None
    )



import factory
from django.utils.text import slugify
from faker import Faker

from machina.core.db.models import get_model


faker = Faker()

Forum = get_model('forum', 'Forum')


class ForumFactory(factory.DjangoModelFactory):
    name = faker.text(max_nb_chars=100)
    slug = factory.LazyAttribute(lambda t: slugify(t.name))

    # Link forum specific
    link = faker.uri()

    class Meta:
        model = Forum


def build_forum(**attrs):
    """Create a new forum but do not save it."""
    params_dict = {'type': Forum.FORUM_POST}
    params_dict.update(attrs)
    forum = ForumFactory.build(**params_dict)
    return forum


def create_forum(**attrs):
    """Save a new forum."""
    forum = build_forum(**attrs)
    forum.save()
    return forum


def build_category_forum(**attrs):
    """Create a new category forum but do not save it."""
    params_dict = {'type': Forum.FORUM_CAT}
    params_dict.update(attrs)
    category = ForumFactory.build(**params_dict)
    return category


def create_category_forum(**attrs):
    """Save a new category forum."""
    category = build_category_forum(**attrs)
    category.save()
    return category


def build_link_forum(**attrs):
    """Create a new link forum but do not save it."""
    params_dict = {'type': Forum.FORUM_LINK}
    params_dict.update(attrs)
    link = ForumFactory.build(**params_dict)
    return link


def create_link_forum(**attrs):
    """Save a new link forum."""
    link = build_link_forum(**attrs)
    link.save()
    return link



"""
    Forum feeds
    ===========

    This module defines the feed abstractions provided by the forum application. It allows to
    generate RSS feeds.

"""

from django.contrib.syndication.views import Feed
from django.shortcuts import get_object_or_404
from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy as _

from machina.core.db.models import get_model


Forum = get_model('forum', 'Forum')
Topic = get_model('forum_conversation', 'Topic')


class LastTopicsFeed(Feed):
    """ Provides feed items for the latest forum topics. """

    # Standard RSS elements
    title = _('Latest topics')
    description = _('Latest topics updated on the forums')
    link = reverse_lazy('forum:index')

    # Item elements
    title_template = 'forum_feeds/topics_title.html'
    description_template = 'forum_feeds/topics_description.html'

    def get_object(self, request, *args, **kwargs):
        """ Handles the considered object. """
        forum_pk = kwargs.get('forum_pk', None)
        descendants = kwargs.get('descendants', None)
        self.user = request.user

        if forum_pk:
            forum = get_object_or_404(Forum, pk=forum_pk)
            forums_qs = (
                forum.get_descendants(include_self=True)
                if descendants else Forum.objects.filter(pk=forum_pk)
            )
            self.forums = request.forum_permission_handler.get_readable_forums(
                forums_qs, request.user,
            )
        else:
            self.forums = request.forum_permission_handler.get_readable_forums(
                Forum.objects.all(), request.user,
            )

    def items(self):
        """ Returns the items to include into the feed. """
        return Topic.objects.filter(forum__in=self.forums, approved=True).order_by('-last_post_on')

    def item_link(self, item):
        """ Generates a link for a specific item of the feed. """
        return reverse_lazy(
            'forum_conversation:topic',
            kwargs={
                'forum_slug': item.forum.slug,
                'forum_pk': item.forum.pk,
                'slug': item.slug,
                'pk': item.id,
            },
        )

    def item_pubdate(self, item):
        """ Returns the item's creation date or update date. """
        return item.created



import factory
from django.utils.text import slugify
from factory import fuzzy
from faker import Faker

from machina.core.db.models import get_model
from machina.test.factories.auth import UserFactory
from machina.test.factories.forum import ForumFactory


faker = Faker()

Post = get_model('forum_conversation', 'Post')
Topic = get_model('forum_conversation', 'Topic')


class TopicFactory(factory.DjangoModelFactory):
    forum = factory.SubFactory(ForumFactory)
    poster = factory.SubFactory(UserFactory)
    status = Topic.TOPIC_UNLOCKED
    subject = factory.LazyAttribute(lambda t: faker.text(max_nb_chars=200))
    slug = factory.LazyAttribute(lambda t: slugify(t.subject))

    class Meta:
        model = Topic


class PostFactory(factory.DjangoModelFactory):
    topic = factory.SubFactory(TopicFactory)
    poster = factory.SubFactory(UserFactory)
    subject = factory.LazyAttribute(lambda t: faker.text(max_nb_chars=200))
    content = fuzzy.FuzzyText(length=255)

    class Meta:
        model = Post


def build_topic(**attrs):
    """Create a new unlocked topic but do not save it."""
    params_dict = {'type': Topic.TOPIC_POST}
    params_dict.update(attrs)
    topic = TopicFactory.build(**params_dict)
    return topic


def create_topic(**attrs):
    """Save a new unlocked topic."""
    topic = build_topic(**attrs)
    topic.save()
    return topic



"""
    Forum tracking URLs
    ===================

    This module defines URL patterns associated with the django-machina's ``forum_tracking``
    application.

"""

from django.urls import path
from django.utils.translation import ugettext_lazy as _

from machina.core.loading import get_class
from machina.core.urls import URLPatternsFactory


class ForumTrackingURLPatternsFactory(URLPatternsFactory):
    """ Allows to generate the URL patterns of the ``forum_search`` application. """

    app_namespace = 'forum_tracking'

    mark_forums_read_view = get_class('forum_tracking.views', 'MarkForumsReadView')
    mark_topics_read_view = get_class('forum_tracking.views', 'MarkTopicsReadView')
    unread_topics_view = get_class('forum_tracking.views', 'UnreadTopicsView')

    def get_urlpatterns(self):
        """ Returns the URL patterns managed by the considered factory / application. """
        return [
            path(
                _('mark/forums/'),
                self.mark_forums_read_view.as_view(),
                name='mark_all_forums_read',
            ),
            path(
                _('mark/forums/<int:pk>/'),
                self.mark_forums_read_view.as_view(),
                name='mark_subforums_read',
            ),
            path(
                _('mark/forum/<int:pk>/topics/'),
                self.mark_topics_read_view.as_view(),
                name='mark_topics_read',
            ),
            path(
                _('unread-topics/'),
                self.unread_topics_view.as_view(),
                name='unread_topics',
            ),
        ]


urlpatterns_factory = ForumTrackingURLPatternsFactory()



"""
    Django-machina
    ==============

    Django-machina is a Django forum engine for building powerful community driven websites. It
    offers a full-featured yet very extensible forum solution.

"""

import os


__version__ = '1.0.1'


# Main Machina static directory.
MACHINA_MAIN_STATIC_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'static/machina/build',
)

# Main Machina template directory.
MACHINA_MAIN_TEMPLATE_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'templates/machina',
)



"""
    Forum search app config
    =======================

    This module contains the application configuration class - available in the Django app registry.
    For more information on this file, see https://docs.djangoproject.com/en/dev/ref/applications/

"""

from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class ForumSearchAppConfig(AppConfig):
    label = 'forum_search'
    name = 'machina.apps.forum_search'
    verbose_name = _('Machina: Forum searches')



"""
    Forum app config
    ================

    This module contains the application configuration class - available in the Django app registry.
    For more information on this file, see https://docs.djangoproject.com/en/dev/ref/applications/

"""

from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class ForumAppConfig(AppConfig):
    label = 'forum'
    name = 'machina.apps.forum'
    verbose_name = _('Machina: Forum')

    def ready(self):
        """ Executes whatever is necessary when the application is ready. """
        from . import receivers  # noqa: F401



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-06-07 03:00
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forum_permission', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='forumpermission',
            name='codename',
            field=models.CharField(db_index=True, max_length=150, unique=True, verbose_name='Permission codename'),
        ),
        migrations.AlterField(
            model_name='forumpermission',
            name='is_global',
            field=models.BooleanField(db_index=True, default=False, help_text='This permission can be granted globally to all the forums', verbose_name='Global permission'),
        ),
        migrations.AlterField(
            model_name='forumpermission',
            name='is_local',
            field=models.BooleanField(db_index=True, default=True, help_text='This permission can be granted individually for each forum', verbose_name='Local permission'),
        ),
        migrations.AlterField(
            model_name='groupforumpermission',
            name='has_perm',
            field=models.BooleanField(db_index=True, default=True, verbose_name='Has perm'),
        ),
        migrations.AlterField(
            model_name='userforumpermission',
            name='anonymous_user',
            field=models.BooleanField(db_index=True, default=False, verbose_name='Target anonymous user'),
        ),
        migrations.AlterField(
            model_name='userforumpermission',
            name='has_perm',
            field=models.BooleanField(db_index=True, default=True, verbose_name='Has perm'),
        ),
    ]



"""
    Forum conversation managers
    ===========================

    This module defines model managers related to forum conversation models.

"""

from django.db import models


class ApprovedManager(models.Manager):
    def get_queryset(self):
        """ Returns all the approved topics or posts. """
        qs = super().get_queryset()
        qs = qs.filter(approved=True)
        return qs



from django.forms.widgets import Select, Textarea


# Originaly comes from https://djangosnippets.org/snippets/2453/
class SelectWithDisabled(Select):
    """ Subclass of Django's select widget that allows disabling options.

    To disable an option, pass a dict instead of a string for its label, of the form:

        {'label': 'option label', 'disabled': True}

    """

    def create_option(self, name, value, label, selected, index, subindex=None, attrs=None):
        disabled = False
        if isinstance(label, dict):
            label, disabled = label['label'], label['disabled']
        option_dict = super().create_option(
            name, value, label, selected, index, subindex=subindex, attrs=attrs
        )
        if disabled:
            option_dict['attrs']['disabled'] = 'disabled'
        return option_dict


class MarkdownTextareaWidget(Textarea):
    """ A simple Textarea widget using the easymde JS library to provide Markdown editor. """

    class Media:
        css = {
            'all': ('machina/build/css/vendor/easymde.min.css', ),
        }
        js = (
            'machina/build/js/vendor/easymde.min.js',
            'machina/build/js/machina.editor.min.js',
        )

    def render(self, name, value, attrs=None, **kwargs):
        attrs = {} if attrs is None else attrs
        classes = attrs.get('classes', '')
        attrs['class'] = classes + ' machina-mde-markdown'
        return super().render(name, value, attrs, **kwargs)



"""
    Forum search forms
    ==================

    This module defines forms provided by the ``forum_search`` application.

"""

from django import forms
from django.utils.translation import ugettext_lazy as _
from haystack.forms import FacetedSearchForm
from haystack.inputs import AutoQuery

from machina.core.db.models import get_model
from machina.core.loading import get_class


Forum = get_model('forum', 'Forum')

PermissionHandler = get_class('forum_permission.handler', 'PermissionHandler')


class SearchForm(FacetedSearchForm):
    """ Allows to search forum topics and posts. """

    search_topics = forms.BooleanField(label=_('Search only in topic subjects'), required=False)

    search_poster_name = forms.CharField(
        label=_('Search for poster'),
        help_text=_('Enter a user name to limit the search to a specific user.'),
        max_length=255, required=False,
    )

    search_forums = forms.MultipleChoiceField(
        label=_('Search in specific forums'),
        help_text=_('Select the forums you wish to search in.'),
        required=False,
    )

    def __init__(self, *args, **kwargs):
        user = kwargs.pop('user', None)

        super().__init__(*args, **kwargs)

        # Update some fields
        self.fields['q'].label = _('Search for keywords')
        self.fields['q'].widget.attrs['placeholder'] = _('Keywords or phrase')
        self.fields['search_poster_name'].widget.attrs['placeholder'] = _('Poster name')

        self.allowed_forums = PermissionHandler().get_readable_forums(Forum.objects.all(), user)
        if self.allowed_forums:
            self.fields['search_forums'].choices = [
                (f.id, '{} {}'.format('-' * f.margin_level, f.name)) for f in self.allowed_forums
            ]
        else:
            # The user cannot view any single forum, the 'search_forums' field can be deleted
            del self.fields['search_forums']

    def search(self):
        sqs = super().search()

        if not self.is_valid():
            return self.no_query_found()

        # Handles topic-based searches
        if self.cleaned_data['search_topics']:
            sqs = sqs.filter(topic_subject=AutoQuery(self.cleaned_data['q']))

        # Handles searches by poster name
        if self.cleaned_data['search_poster_name']:
            sqs = sqs.filter(poster_name__icontains=self.cleaned_data['search_poster_name'])

        # Handles searches in specific forums if necessary
        if 'search_forums' in self.cleaned_data and self.cleaned_data['search_forums']:
            sqs = sqs.filter(forum__in=self.cleaned_data['search_forums'])
        else:
            forum_ids = self.allowed_forums.values_list('id', flat=True)
            sqs = sqs.filter(forum__in=forum_ids) if forum_ids else sqs.none()

        return sqs



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-03 02:50
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('forum_conversation', '0006_post_enable_signature'),
    ]

    operations = [
        migrations.AddField(
            model_name='topic',
            name='first_post',
            field=models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='forum_conversation.Post', verbose_name='Last post'),
        ),
        migrations.AddField(
            model_name='topic',
            name='last_post',
            field=models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='forum_conversation.Post', verbose_name='Last post'),
        ),
    ]



"""
    Forum member abstract models
    ============================

    This module defines abstract models provided by the ``forum_member`` application.

"""

import os
import uuid

from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _

from machina.conf import settings as machina_settings
from machina.core import validators
from machina.models.fields import ExtendedImageField, MarkupTextField


def get_profile_avatar_upload_to(instance, filename):
    """ Returns a valid upload path for the avatar associated with a forum profile. """
    return instance.get_avatar_upload_to(filename)


class AbstractForumProfile(models.Model):
    """ Represents the profile associated with each forum user. """

    user = models.OneToOneField(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name='forum_profile',
        verbose_name=_('User'),
    )

    # The user's avatar.
    avatar = ExtendedImageField(
        null=True, blank=True, upload_to=get_profile_avatar_upload_to, verbose_name=_('Avatar'),
        **machina_settings.DEFAULT_AVATAR_SETTINGS
    )

    # The user's signature.
    signature = MarkupTextField(
        verbose_name=_('Signature'), blank=True, null=True,
        validators=[
            validators.NullableMaxLengthValidator(machina_settings.PROFILE_SIGNATURE_MAX_LENGTH),
        ],
    )

    # The amount of posts the user has posted (only approved posts are considered here).
    posts_count = models.PositiveIntegerField(verbose_name=_('Total posts'), blank=True, default=0)

    class Meta:
        abstract = True
        app_label = 'forum_member'
        verbose_name = _('Forum profile')
        verbose_name_plural = _('Forum profiles')

    def __str__(self):
        return self.user.username

    def get_avatar_upload_to(self, filename):
        """ Returns the path to upload the associated avatar to. """
        dummy, ext = os.path.splitext(filename)
        return os.path.join(
            machina_settings.PROFILE_AVATAR_UPLOAD_TO,
            '{id}{ext}'.format(id=str(uuid.uuid4()).replace('-', ''), ext=ext),
        )



"""
    Forum signals
    =============

    This modules defines Django signals that can be triggered by the ``forum`` application.

"""

import django.dispatch


forum_moved = django.dispatch.Signal(providing_args=["previous_parent", ])
forum_viewed = django.dispatch.Signal(providing_args=["forum", "user", "request", "response", ])



"""
    Forum permission handler
    ========================

    This module defines a ``PermissionHandler`` abstraction that allows to implement filter or
    access logic related to forums.

"""

import collections
import datetime as dt
from functools import reduce

from django.contrib.auth import get_user_model
from django.db import models
from django.utils.timezone import now
from mptt.utils import get_cached_trees

from machina.conf import settings as machina_settings
from machina.core.db.models import get_model
from machina.core.loading import get_class


Forum = get_model('forum', 'Forum')
GroupForumPermission = get_model('forum_permission', 'GroupForumPermission')
Post = get_model('forum_conversation', 'Post')
TopicPollVote = get_model('forum_polls', 'TopicPollVote')
UserForumPermission = get_model('forum_permission', 'UserForumPermission')

ForumPermissionChecker = get_class('forum_permission.checker', 'ForumPermissionChecker')

get_anonymous_user_forum_key = get_class(
    'forum_permission.shortcuts', 'get_anonymous_user_forum_key')


class PermissionHandler:
    """ Defines filter / access logic related to forums.

    The ``PermissionHandler`` class allows to filter lists of forums and to perform permission
    verifications on forums. It uses the ``ForumPermissionChecker`` class to perform these
    verifications.

    """

    def __init__(self):
        # This dictionary will store the forums that are granted for a specific lit of permission
        # codenames and a given user.
        self._granted_forums_cache = {}

        # This one will store the ancestors of the forums that will be used to perform permission
        # checks.
        self._forum_ancestors_cache = {}

        # This one will store ForumPermissionChecker instances for users whose permissions will be
        # checked.
        self._user_perm_checkers_cache = {}

    # Filtering methods
    # --

    def forum_list_filter(self, qs, user):
        """ Filters the given queryset in order to return a list of forums that can be seen and read
            by the specified user (at least).
        """
        # Any superuser should see all the forums
        if user.is_superuser:
            return qs

        # Check whether the forums can be viewed by the given user
        forums_to_hide = self._get_hidden_forum_ids(qs, user)

        return qs.exclude(id__in=forums_to_hide)

    def get_readable_forums(self, forums, user):
        """ Returns a queryset of forums that can be read by the considered user. """
        # Any superuser should be able to read all the forums.
        if user.is_superuser:
            return forums

        # Fetches the forums that can be read by the given user.
        readable_forums = self._get_forums_for_user(
            user, ['can_read_forum', ], use_tree_hierarchy=True)
        return forums.filter(id__in=[f.id for f in readable_forums]) \
            if isinstance(forums, (models.Manager, models.QuerySet)) \
            else list(filter(lambda f: f in readable_forums, forums))

    # Verification methods
    # --

    # Forums

    def can_read_forum(self, forum, user):
        """ Given a forum, checks whether the user can read its content. """
        return self._perform_basic_permission_check(forum, user, 'can_read_forum')

    # Posts and topics

    def can_add_topic(self, forum, user):
        """ Given a forum, checks whether the user can append topics to it. """
        return self._perform_basic_permission_check(forum, user, 'can_start_new_topics')

    def can_add_stickies(self, forum, user):
        """ Given a forum, checks whether the user can append stickies to it. """
        return self._perform_basic_permission_check(forum, user, 'can_post_stickies')

    def can_add_announcements(self, forum, user):
        """ Given a forum, checks whether the user can append announcements to it. """
        return self._perform_basic_permission_check(forum, user, 'can_post_announcements')

    def can_post_without_approval(self, forum, user):
        """ Given a forum, checks whether the user can add a posts and topics without approval. """
        return self._perform_basic_permission_check(forum, user, 'can_post_without_approval')

    def can_add_post(self, topic, user):
        """ Given a topic, checks whether the user can append posts to it. """
        can_add_post = self._perform_basic_permission_check(
            topic.forum, user, 'can_reply_to_topics',
        )
        can_add_post &= (
            not topic.is_locked or
            self._perform_basic_permission_check(topic.forum, user, 'can_reply_to_locked_topics')
        )
        return can_add_post

    def can_edit_post(self, post, user):
        """ Given a forum post, checks whether the user can edit the latter. """
        checker = self._get_checker(user)

        # A user can edit a post if...
        #     they are a superuser
        #     they are the original poster of the forum post
        #     they belong to the forum moderators
        is_author = self._is_post_author(post, user)
        can_edit = (
            user.is_superuser or
            (
                is_author and checker.has_perm('can_edit_own_posts', post.topic.forum) and
                not post.topic.is_locked
            ) or
            checker.has_perm('can_edit_posts', post.topic.forum)
        )
        return can_edit

    def can_delete_post(self, post, user):
        """
        Given a forum post, checks whether the user can delete the latter.
        """
        checker = self._get_checker(user)

        # A user can delete a post if...
        #     they are a superuser
        #     they are the original poster of the forum post
        #     they belong to the forum moderators
        is_author = self._is_post_author(post, user)
        can_delete = (
            user.is_superuser or
            (is_author and checker.has_perm('can_delete_own_posts', post.topic.forum)) or
            checker.has_perm('can_delete_posts', post.topic.forum)
        )
        return can_delete

    # Polls

    def can_create_polls(self, forum, user):
        """ Given a forum, checks whether the user can add a topic with an embedded poll. """
        return self._perform_basic_permission_check(forum, user, 'can_create_polls')

    def can_vote_in_poll(self, poll, user):
        """ Given a poll, checks whether the user can answer to it. """
        # First we have to check if the poll is curently open
        if poll.duration:
            poll_dtend = poll.created + dt.timedelta(days=poll.duration)
            if poll_dtend < now():
                return False

        # Is this user allowed to vote in polls in the current forum?
        can_vote = (
            self._perform_basic_permission_check(poll.topic.forum, user, 'can_vote_in_polls') and
            not poll.topic.is_locked
        )

        # Retrieve the user votes for the considered poll
        user_votes = TopicPollVote.objects.filter(poll_option__poll=poll)
        if user.is_anonymous:
            forum_key = get_anonymous_user_forum_key(user)
            if forum_key:
                user_votes = user_votes.filter(anonymous_key=forum_key)
            else:
                # If the forum key of the anonymous user cannot be retrieved, the user should not be
                # allowed to vote in the considered poll.
                user_votes = user_votes.none()
                can_vote = False
        else:
            user_votes = user_votes.filter(voter=user)

        # If the user has already voted, they can vote again if the vote changes are allowed
        if user_votes.exists() and can_vote:
            can_vote = poll.user_changes

        return can_vote

    # Attachments

    def can_attach_files(self, forum, user):
        """ Given a forum, checks whether the user can add attachments to posts. """
        return self._perform_basic_permission_check(forum, user, 'can_attach_file')

    def can_download_files(self, forum, user):
        """ Given a forum, checks whether the user can download files attached to posts. """
        return self._perform_basic_permission_check(forum, user, 'can_download_file')

    # Topic subscription

    def can_subscribe_to_topic(self, topic, user):
        """ Given a topic, checks whether the user can add it to their subscription list. """
        # A user can subscribe to topics if they are authenticated and if they have the permission
        # to read the related forum. Of course a user can subscribe only if they have not already
        # subscribed to the considered topic.
        return (
            user.is_authenticated and
            not topic.has_subscriber(user) and
            self._perform_basic_permission_check(topic.forum, user, 'can_read_forum')
        )

    def can_unsubscribe_from_topic(self, topic, user):
        """ Given a topic, checks whether the user can remove it from their subscription list. """
        # A user can unsubscribe from topics if they are authenticated and if they have the
        # permission to read the related forum. Of course a user can unsubscribe only if they are
        #  already a subscriber of the considered topic.
        return (
            user.is_authenticated and
            topic.has_subscriber(user) and
            self._perform_basic_permission_check(topic.forum, user, 'can_read_forum')
        )

    # Moderation

    def get_moderation_queue_forums(self, user):
        """ Returns the list of forums whose posts can be approved by the considered user. """
        return self._get_forums_for_user(user, ['can_approve_posts', ])

    def can_access_moderation_queue(self, user):
        """ Returns True if the passed user can access the moderation queue. """
        return len(self.get_moderation_queue_forums(user)) > 0

    def can_lock_topics(self, forum, user):
        """ Given a forum, checks whether the user can lock its topics. """
        return self._perform_basic_permission_check(forum, user, 'can_lock_topics')

    def can_move_topics(self, forum, user):
        """ Given a forum, checks whether the user can move its topics to another forum. """
        return self._perform_basic_permission_check(forum, user, 'can_move_topics')

    def get_target_forums_for_moved_topics(self, user):
        """ Returns a list of forums in which the considered user can add topics that have been
            moved from another forum.
        """
        return [f for f in self._get_forums_for_user(user, ['can_move_topics', ]) if f.is_forum]

    def can_delete_topics(self, forum, user):
        """ Given a forum, checks whether the user can delete its topics.

        Note: the ``can_delete_posts`` permission is used here because a user who can delete all the
        posts of a topic is also able to delete the topic itself.

        """
        return self._perform_basic_permission_check(forum, user, 'can_delete_posts')

    def can_update_topics_to_normal_topics(self, forum, user):
        """ Given a forum, checks whether the user can change its topic types to normal topics. """
        return self._perform_basic_permission_check(forum, user, 'can_edit_posts')

    def can_update_topics_to_sticky_topics(self, forum, user):
        """ Given a forum, checks whether the user can change its topic types to sticky topics. """
        return (
            self._perform_basic_permission_check(forum, user, 'can_edit_posts') and
            self._perform_basic_permission_check(forum, user, 'can_post_stickies')
        )

    def can_update_topics_to_announces(self, forum, user):
        """ Given a forum, checks whether the user can change its topic types to announces. """
        return (
            self._perform_basic_permission_check(forum, user, 'can_edit_posts') and
            self._perform_basic_permission_check(forum, user, 'can_post_announcements')
        )

    def can_approve_posts(self, forum, user):
        """ Given a forum, checks whether the user can approve its posts. """
        return self._perform_basic_permission_check(forum, user, 'can_approve_posts')

    # Common
    # --

    def _is_post_author(self, post, user):
        return (
            (post.poster == user) if user.is_authenticated else
            (
                post.anonymous_key is not None and
                post.anonymous_key == get_anonymous_user_forum_key(user)
            )
        )

    def _get_hidden_forum_ids(self, forums, user):
        """ Given a set of forums and a user, returns the list of forums that are not visible by
            this user.
        """
        visible_forums = self._get_forums_for_user(
            user, ['can_see_forum', 'can_read_forum', ], use_tree_hierarchy=True,
        )
        return forums.exclude(id__in=[f.id for f in visible_forums])

    def _get_forums_for_user(self, user, perm_codenames, use_tree_hierarchy=False):
        """ Returns all the forums that satisfy the given list of permission codenames.

        User and group forum permissions are used.

        If the ``use_tree_hierarchy`` keyword argument is set the granted forums will be filtered so
        that a forum which has an ancestor which is not in the granted forums set will not be
        returned.
        """
        granted_forums_cache_key = '{}__{}'.format(
            ':'.join(perm_codenames), user.id if not user.is_anonymous else 'anonymous',
        )

        if granted_forums_cache_key in self._granted_forums_cache:
            return self._granted_forums_cache[granted_forums_cache_key]

        forums = self._get_all_forums()

        # First check if the user is a superuser and if so, returns the forum queryset immediately.
        if user.is_superuser:  # pragma: no cover
            forum_objects = forums

        else:
            # Generates the appropriate queryset filter in order to handle both authenticated users
            # and anonymous users.
            user_kwargs_filter = {'anonymous_user': True} if user.is_anonymous else {'user': user}

            # Get all the user permissions for the considered user.
            user_perms = (
                UserForumPermission.objects
                .filter(**user_kwargs_filter)
                .filter(permission__codename__in=perm_codenames)
            )

            # The first thing to do is to compute three lists of permissions: one containing only
            # globally granted permissions, one containing granted permissions (these permissions
            # are associated with specific forums) and one containing non granted permissions (the
            # latest are also associated with specific forums).
            globally_granted_user_perms = list(
                filter(lambda p: p.has_perm and p.forum_id is None, user_perms)
            )
            per_forum_granted_user_perms = list(
                filter(lambda p: p.has_perm and p.forum_id is not None, user_perms)
            )
            per_forum_nongranted_user_perms = list(
                filter(lambda p: not p.has_perm and p.forum_id is not None, user_perms)
            )

            # Using the previous lists we are able to compute a list of forums ids for which
            # permissions are explicitly not granted. It should be noted that any permission that is
            # explicitely set for a user will not be considered as non granted if a "non granted"
            # permission also exists. The explicitly granted permissions always win precedence.
            granted_user_forum_ids = [p.forum_id for p in per_forum_granted_user_perms]
            nongranted_forum_ids = [
                p.forum_id for p in per_forum_nongranted_user_perms
                if p.forum_id not in granted_user_forum_ids
            ]

            required_perm_codenames_count = len(perm_codenames)
            initial_forum_ids = [f.id for f in forums]

            # Now we build a dictionary allowing to associate each forum ID of the initial queryset
            # with a set of permissions that are granted for the considered forum.
            granted_permissions_per_forum = collections.defaultdict(set)
            for perm in per_forum_granted_user_perms:
                granted_permissions_per_forum[perm.forum_id].add(perm.permission_id)
            for forum_id in initial_forum_ids:
                granted_permissions_per_forum[forum_id].update(
                    [perm.permission_id for perm in globally_granted_user_perms]
                )

            if not user.is_anonymous:
                user_model = get_user_model()

                # Get all the group permissions for the considered user.
                group_perms = (
                    GroupForumPermission.objects
                    .filter(
                        **{
                            'group__{0}'.format(user_model.groups.field.related_query_name()): user
                        }
                    )
                    .filter(permission__codename__in=perm_codenames)
                )

                # Again, we compute three lists of permissions. But this time we are considering
                # group permissions. The first list contains only globally granted permissions. The
                # second one contains only granted permissions that are associated with specific
                # forums. The third list contains non granted permissions.
                globally_granted_group_perms = list(
                    filter(lambda p: p.has_perm and p.forum_id is None, group_perms)
                )
                per_forum_granted_group_perms = list(
                    filter(lambda p: p.has_perm and p.forum_id is not None, group_perms)
                )
                per_forum_nongranted_group_perms = list(
                    filter(lambda p: not p.has_perm and p.forum_id is not None, group_perms)
                )

                # Now we can update the list of forums ids for which permissions are explicitly not
                # granted.
                granted_group_forum_ids = [p.forum_id for p in per_forum_granted_group_perms]
                nongranted_forum_ids += [
                    p.forum_id for p in per_forum_nongranted_group_perms
                    if p.forum_id not in granted_group_forum_ids
                ]

                # Now we will update our previous dictionary that associated each forum ID with a
                # set of granted permissions (at the user level). We will update it with the new
                # permissions we've computed for the user's groups.
                for perm in per_forum_granted_group_perms:
                    granted_permissions_per_forum[perm.forum_id].add(perm.permission_id)
                for forum_id in initial_forum_ids:
                    granted_permissions_per_forum[forum_id].update(
                        [perm.permission_id for perm in globally_granted_group_perms]
                    )

            # We keep only the forum IDs for which the length of the set containing the granted
            # permissions is equal to the number of initial permission codenames. The other forums
            # have not all the required granted permissions, so we just throw them away.
            for forum_id in list(granted_permissions_per_forum):
                if len(granted_permissions_per_forum[forum_id]) < required_perm_codenames_count:
                    del granted_permissions_per_forum[forum_id]

            # Alright! It is now possible to filter the initial queryset using the forums associated
            # with the granted permissions and the list of forums for which permissions are
            # explicitly not granted.
            forum_objects = [
                f for f in forums
                if f.id in granted_permissions_per_forum and f.id not in nongranted_forum_ids
            ]

            if (
                not user.is_anonymous and
                set(perm_codenames).issubset(
                    set(machina_settings.DEFAULT_AUTHENTICATED_USER_FORUM_PERMISSIONS)
                )
            ):
                forum_objects += [
                    f for f in forums if f.id not in nongranted_forum_ids and f not in forum_objects
                ]

            if use_tree_hierarchy:
                forum_objects = self._filter_granted_forums_using_tree(forum_objects)

        self._granted_forums_cache[granted_forums_cache_key] = forum_objects
        return forum_objects

    def _filter_granted_forums_using_tree(self, granted_forums):
        top_nodes = self._get_top_nodes()
        d = reduce(
            lambda a, b: a + self._filter_granted_node_using_tree(b, granted_forums), top_nodes, []
        )
        return Forum.objects.filter(id__in=d)

    def _filter_granted_node_using_tree(self, f, granted_forums):
        if f in granted_forums:
            return [f.id, ] + reduce(
                lambda a, b: a + self._filter_granted_node_using_tree(b, granted_forums),
                f.get_children(), []
            )
        return []

    def _get_top_nodes(self):
        if not hasattr(self, '_top_nodes'):
            self._top_nodes = get_cached_trees(Forum.objects.all())
        return self._top_nodes

    def _perform_basic_permission_check(self, forum, user, permission):
        """ Given a forum and a user, checks whether the latter has the passed permission.

        The workflow is:

            1. The permission is granted if the user is a superuser
            2. If not, a check is performed with the given permission

        """
        checker = self._get_checker(user)

        # The action is granted if...
        #     the user is the superuser
        #     the user has the permission to do so
        check = (user.is_superuser or checker.has_perm(permission, forum))
        return check

    def _get_checker(self, user):
        """ Return a ForumPermissionChecker instance for the given user. """
        user_perm_checkers_cache_key = user.id if not user.is_anonymous else 'anonymous'

        if user_perm_checkers_cache_key in self._user_perm_checkers_cache:
            return self._user_perm_checkers_cache[user_perm_checkers_cache_key]

        checker = ForumPermissionChecker(user)
        self._user_perm_checkers_cache[user_perm_checkers_cache_key] = checker
        return checker

    def _get_all_forums(self):
        """ Returns all forums. """
        if not hasattr(self, '_all_forums'):
            self._all_forums = list(Forum.objects.all())
        return self._all_forums



from django import template

from machina.core.loading import get_class


TrackingHandler = get_class('forum_tracking.handler', 'TrackingHandler')

register = template.Library()


@register.simple_tag(takes_context=True)
def get_unread_topics(context, topics, user):
    """ This will return a list of unread topics for the given user from a given set of topics.

    Usage::

        {% get_unread_topics topics request.user as unread_topics %}

    """
    request = context.get('request', None)
    return TrackingHandler(request=request).get_unread_topics(topics, user)



"""
    Forum attachments models
    ========================

    This module defines models provided by the ``forum_attachments`` application.

"""

from machina.apps.forum_conversation.forum_attachments.abstract_models import AbstractAttachment
from machina.core.db.models import model_factory


Attachment = model_factory(AbstractAttachment)



"""
    Forum tracking views
    ====================

    This module defines views provided by the ``forum_tracking`` application.

"""

from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from django.views.generic import ListView, TemplateView
from django.views.generic.detail import BaseDetailView, SingleObjectTemplateResponseMixin

from machina.conf import settings as machina_settings
from machina.core.db.models import get_model
from machina.core.loading import get_class, get_classes


Forum = get_model('forum', 'Forum')
ForumReadTrack, TopicReadTrack = get_classes(
    'forum_tracking.models', ['ForumReadTrack', 'TopicReadTrack']
)
Topic = get_model('forum_conversation', 'Topic')

TrackingHandler = get_class('forum_tracking.handler', 'TrackingHandler')
track_handler = TrackingHandler()

PermissionRequiredMixin = get_class('forum_permission.viewmixins', 'PermissionRequiredMixin')


class MarkForumsReadView(LoginRequiredMixin, TemplateView):
    """ Marks a set of forums as read. """

    success_message = _('Forums have been marked read.')
    template_name = 'forum_tracking/mark_forums_read.html'

    def get(self, request, pk=None):
        """ Handles GET requests. """
        self.top_level_forum = get_object_or_404(Forum, pk=pk) if pk else None
        return super().get(request, pk)

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        context['top_level_forum'] = self.top_level_forum
        context['top_level_forum_url'] = self.get_top_level_forum_url()
        return context

    def get_top_level_forum_url(self):
        """ Returns the parent forum from which forums are marked as read. """
        return (
            reverse('forum:index') if self.top_level_forum is None else
            reverse(
                'forum:forum',
                kwargs={'slug': self.top_level_forum.slug, 'pk': self.kwargs['pk']},
            )
        )

    def mark_as_read(self, request, pk):
        """ Marks the considered forums as read. """
        if self.top_level_forum is not None:
            forums = request.forum_permission_handler.get_readable_forums(
                self.top_level_forum.get_descendants(include_self=True), request.user,
            )
        else:
            forums = request.forum_permission_handler.get_readable_forums(
                Forum.objects.all(), request.user,
            )

        # Marks forums as read
        track_handler.mark_forums_read(forums, request.user)

        if len(forums):
            messages.success(request, self.success_message)

        return HttpResponseRedirect(self.get_top_level_forum_url())

    def post(self, request, pk=None):
        """ Handles POST requests. """
        self.top_level_forum = get_object_or_404(Forum, pk=pk) if pk else None
        return self.mark_as_read(request, pk)


class MarkTopicsReadView(
    LoginRequiredMixin, PermissionRequiredMixin, SingleObjectTemplateResponseMixin, BaseDetailView,
):
    """ Marks a set of topics as read. """

    model = Forum
    permission_required = ['can_read_forum', ]
    success_message = _('Topics have been marked read.')
    template_name = 'forum_tracking/mark_topics_read.html'

    def get(self, request, pk):
        """ Handles GET requests. """
        self.forum = get_object_or_404(Forum, pk=pk)
        return super().get(request, pk)

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        context['forum_url'] = self.get_forum_url()
        return context

    def get_controlled_object(self):
        """ Return the considered forum in order to allow permission checks. """
        return Forum.objects.get(pk=self.kwargs['pk'])

    def get_forum_url(self):
        """ Returns the url of the forum whose topics will be marked read. """
        return reverse('forum:forum', kwargs={'slug': self.forum.slug, 'pk': self.forum.pk})

    def mark_topics_read(self, request, pk):
        """ Marks forum topics as read. """
        track_handler.mark_forums_read([self.forum, ], request.user)
        messages.success(request, self.success_message)
        return HttpResponseRedirect(self.get_forum_url())

    def post(self, request, pk):
        """ Handles POST requests. """
        self.forum = get_object_or_404(Forum, pk=pk)
        return self.mark_topics_read(request, pk)


class UnreadTopicsView(LoginRequiredMixin, ListView):
    """ Displays unread topics for the current user. """

    context_object_name = 'topics'
    paginate_by = machina_settings.FORUM_TOPICS_NUMBER_PER_PAGE
    template_name = 'forum_tracking/unread_topic_list.html'

    def get_queryset(self):
        """ Returns the list of items for this view. """
        forums = self.request.forum_permission_handler.get_readable_forums(
            Forum.objects.all(), self.request.user,
        )
        topics = Topic.objects.filter(forum__in=forums)
        topics_pk = map(lambda t: t.pk, track_handler.get_unread_topics(topics, self.request.user))
        return Topic.approved_objects.filter(pk__in=topics_pk).order_by('-last_post_on')



from django.urls import reverse


class AdminBaseViewTestMixin:
    """
    Mixin that can be used to append a test to an AdminClientTestCase in order to
    test that the admin base views associated with a given model are accessible
    and so avoid some boilerplate code.
    """
    model = None

    def test_has_accessible_base_views(self):
        model = self.model
        urls = (
            'admin:{}_{}_changelist',
            'admin:{}_{}_add',
        )
        try:
            module_name = model._meta.module_name
        except AttributeError:  # pragma: no cover
            module_name = model._meta.model_name
        urls = [raw_url.format(model._meta.app_label, module_name) for raw_url in urls]
        for raw_url in urls:
            url = reverse(raw_url)
            response = self.client.get(url, follow=True)
            assert response.status_code == 200



"""
    Machina URLs
    ============

    This module imports all the URLs defined by the forum-related applications.

"""

from django.conf.urls import include, url
from django.utils.translation import ugettext_lazy as _

from machina.core.loading import get_class
from machina.core.urls import URLPatternsFactory


class BoardURLPatternsFactory(URLPatternsFactory):
    """ Allows to generate the URL patterns of the whole forum application. """

    forum_urlpatterns_factory = get_class('forum.urls', 'urlpatterns_factory')
    conversation_urlpatterns_factory = get_class('forum_conversation.urls', 'urlpatterns_factory')
    feeds_urlpatterns_factory = get_class('forum_feeds.urls', 'urlpatterns_factory')
    member_urlpatterns_factory = get_class('forum_member.urls', 'urlpatterns_factory')
    moderation_urlpatterns_factory = get_class('forum_moderation.urls', 'urlpatterns_factory')
    search_urlpatterns_factory = get_class('forum_search.urls', 'urlpatterns_factory')
    tracking_urlpatterns_factory = get_class('forum_tracking.urls', 'urlpatterns_factory')

    def get_urlpatterns(self):
        """ Returns the URL patterns managed by the considered factory / application. """
        return [
            url(r'', include(self.forum_urlpatterns_factory.urlpatterns)),
            url(r'', include(self.conversation_urlpatterns_factory.urlpatterns)),
            url(_(r'^feeds/'), include(self.feeds_urlpatterns_factory.urlpatterns)),
            url(_(r'^member/'), include(self.member_urlpatterns_factory.urlpatterns)),
            url(_(r'^moderation/'), include(self.moderation_urlpatterns_factory.urlpatterns)),
            url(_(r'^search/'), include(self.search_urlpatterns_factory.urlpatterns)),
            url(_(r'^tracking/'), include(self.tracking_urlpatterns_factory.urlpatterns)),
        ]


urlpatterns_factory = BoardURLPatternsFactory()
urlpatterns = urlpatterns_factory.get_urlpatterns()



"""
    Forum member signal receivers
    =============================

    This module defines signal receivers.

"""

from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import F
from django.db.models.signals import post_delete, pre_save
from django.dispatch import receiver

from machina.core.db.models import get_model


User = get_user_model()

Post = get_model('forum_conversation', 'Post')
ForumProfile = get_model('forum_member', 'ForumProfile')


@receiver(pre_save, sender=Post)
def increase_posts_count(sender, instance, **kwargs):
    """ Increases the member's post count after a post save.

    This receiver handles the update of the profile related to the user who is the poster of the
    forum post being created or updated.
    """

    if instance.poster is None:
        # An anonymous post is considered. No profile can be updated in
        # that case.
        return

    profile, dummy = ForumProfile.objects.get_or_create(user=instance.poster)
    increase_posts_count = False

    if instance.pk:
        try:
            old_instance = instance.__class__._default_manager.get(pk=instance.pk)
        except ObjectDoesNotExist:  # pragma: no cover
            # This should never happen (except with django loaddata command)
            increase_posts_count = True
            old_instance = None
        if old_instance and old_instance.approved is False and instance.approved is True:
            increase_posts_count = True
    elif instance.approved:
        increase_posts_count = True

    if increase_posts_count:
        profile.posts_count = F('posts_count') + 1
        profile.save()


@receiver(pre_save, sender=Post)
def decrease_posts_count_after_post_unaproval(sender, instance, **kwargs):
    """ Decreases the member's post count after a post unaproval.

    This receiver handles the unaproval of a forum post: the posts count associated with the post's
    author is decreased.
    """
    if not instance.pk:
        # Do not consider posts being created.
        return

    profile, dummy = ForumProfile.objects.get_or_create(user=instance.poster)

    try:
        old_instance = instance.__class__._default_manager.get(pk=instance.pk)
    except ObjectDoesNotExist:  # pragma: no cover
        # This should never happen (except with django loaddata command)
        return

    if old_instance and old_instance.approved is True and instance.approved is False:
        profile.posts_count = F('posts_count') - 1
        profile.save()


@receiver(post_delete, sender=Post)
def decrease_posts_count_after_post_deletion(sender, instance, **kwargs):
    """ Decreases the member's post count after a post deletion.

    This receiver handles the deletion of a forum post: the posts count related to the post's
    author is decreased.
    """
    if not instance.approved:
        # If a post has not been approved, it has not been counted.
        # So do not decrement count
        return

    try:
        assert instance.poster_id is not None
        poster = User.objects.get(pk=instance.poster_id)
    except AssertionError:
        # An anonymous post is considered. No profile can be updated in
        # that case.
        return
    except ObjectDoesNotExist:  # pragma: no cover
        # This can happen if a User instance is deleted. In that case the
        # User instance is not available and the receiver should return.
        return

    profile, dummy = ForumProfile.objects.get_or_create(user=poster)
    if profile.posts_count:
        profile.posts_count = F('posts_count') - 1
        profile.save()



"""
    Forum member models
    ===================

    This module defines models provided by the ``forum_member`` application.

"""

from machina.apps.forum_member.abstract_models import AbstractForumProfile
from machina.core.db.models import model_factory


ForumProfile = model_factory(AbstractForumProfile)



"""
    Forum tracking handler
    ======================

    This module defines a ``TrackingHandler`` abstraction that allows to identify unread forums and
    topics.

"""

from django.db.models import F, Q

from machina.core.db.models import get_model
from machina.core.loading import get_class


Forum = get_model('forum', 'Forum')
ForumReadTrack = get_model('forum_tracking', 'ForumReadTrack')
TopicReadTrack = get_model('forum_tracking', 'TopicReadTrack')

PermissionHandler = get_class('forum_permission.handler', 'PermissionHandler')


class TrackingHandler:
    """ Provides utility methods to compute unread forums and topics.

    The TrackingHandler allows to filter list of forums and list of topics in order to get only the
    forums which contain unread topics or the unread topics.

    """

    def __init__(self, request=None):
        self.request = request
        self.perm_handler = request.forum_permission_handler if request \
            else PermissionHandler()

    def get_unread_forums(self, user):
        """ Returns the list of unread forums for the given user. """
        return self.get_unread_forums_from_list(
            user, self.perm_handler.get_readable_forums(Forum.objects.all(), user))

    def get_unread_forums_from_list(self, user, forums):
        """ Returns the list of unread forums for the given user from a given list of forums. """
        unread_forums = []

        # A user which is not authenticated will never see a forum as unread
        if not user.is_authenticated:
            return unread_forums

        unread = ForumReadTrack.objects.get_unread_forums_from_list(forums, user)
        unread_forums.extend(unread)

        return unread_forums

    def get_unread_topics(self, topics, user):
        """ Returns a list of unread topics for the given user from a given set of topics. """
        unread_topics = []

        # A user which is not authenticated will never see a topic as unread.
        # If there are no topics to consider, we stop here.
        if not user.is_authenticated or topics is None or not len(topics):
            return unread_topics

        # A topic can be unread if a track for itself exists with a mark time that
        # is less important than its update date.
        topic_ids = [topic.id for topic in topics]
        topic_tracks = TopicReadTrack.objects.filter(topic__in=topic_ids, user=user)
        tracked_topics = dict(topic_tracks.values_list('topic__pk', 'mark_time'))

        if tracked_topics:
            for topic in topics:
                topic_last_modification_date = topic.last_post_on or topic.created
                if (
                    topic.id in tracked_topics.keys() and
                    topic_last_modification_date > tracked_topics[topic.id]
                ):
                    unread_topics.append(topic)

        # A topic can be unread if a track for its associated forum exists with
        # a mark time that is less important than its creation or update date.
        forum_ids = [topic.forum_id for topic in topics]
        forum_tracks = ForumReadTrack.objects.filter(forum_id__in=forum_ids, user=user)
        tracked_forums = dict(forum_tracks.values_list('forum__pk', 'mark_time'))

        if tracked_forums:
            for topic in topics:
                topic_last_modification_date = topic.last_post_on or topic.created
                if (
                    (topic.forum_id in tracked_forums.keys() and topic.id not in tracked_topics) and
                    topic_last_modification_date > tracked_forums[topic.forum_id]
                ):
                    unread_topics.append(topic)

        # A topic can be unread if no tracks exists for it
        for topic in topics:
            if topic.forum_id not in tracked_forums and topic.id not in tracked_topics:
                unread_topics.append(topic)

        return list(set(unread_topics))

    def mark_forums_read(self, forums, user):
        """ Marks a list of forums as read. """
        if not forums or not user.is_authenticated:
            return

        forums = sorted(forums, key=lambda f: f.level)

        # Update all forum tracks to the current date for the considered forums
        for forum in forums:
            forum_track = ForumReadTrack.objects.get_or_create(forum=forum, user=user)[0]
            forum_track.save()
        # Delete all the unnecessary topic tracks
        TopicReadTrack.objects.filter(topic__forum__in=forums, user=user).delete()
        # Update parent forum tracks
        self._update_parent_forum_tracks(forums[0], user)

    def mark_topic_read(self, topic, user):
        """ Marks a topic as read. """
        if not user.is_authenticated:
            return

        forum = topic.forum
        try:
            forum_track = ForumReadTrack.objects.get(forum=forum, user=user)
        except ForumReadTrack.DoesNotExist:
            forum_track = None

        if (
            forum_track is None or
            (topic.last_post_on and forum_track.mark_time < topic.last_post_on)
        ):
            topic_track, created = TopicReadTrack.objects.get_or_create(topic=topic, user=user)
            if not created:
                topic_track.save()  # mark_time filled

            # If no other topic is unread inside the considered forum, the latter should also be
            # marked as read.
            unread_topics = (
                forum.topics
                .filter(
                    Q(tracks__user=user, tracks__mark_time__lt=F('last_post_on')) |
                    Q(
                        forum__tracks__user=user, forum__tracks__mark_time__lt=F('last_post_on'),
                        tracks__isnull=True,
                    )
                )
                .exclude(id=topic.id)
            )

            forum_topic_tracks = TopicReadTrack.objects.filter(topic__forum=forum, user=user)
            if (
                not unread_topics.exists() and
                (
                    forum_track is not None or
                    forum_topic_tracks.count() == forum.topics.filter(approved=True).count()
                )
            ):
                # The topics that are marked as read inside the forum for the given user will be
                # deleted while the forum track associated with the user must be created or updated.
                # This is done only if there are as many topic tracks as approved topics in case
                # the related forum has not beem previously marked as read.
                TopicReadTrack.objects.filter(topic__forum=forum, user=user).delete()
                forum_track, _ = ForumReadTrack.objects.get_or_create(forum=forum, user=user)
                forum_track.save()

                # Update parent forum tracks
                self._update_parent_forum_tracks(forum, user)

    def _update_parent_forum_tracks(self, forum, user):
        for forum in forum.get_ancestors(ascending=True):
            # If no other topics are unread inside the considered forum, the latter should also
            # be marked as read.
            unread_topics = (
                forum.topics
                .filter(
                    Q(tracks__user=user, tracks__mark_time__lt=F('last_post_on')) |
                    Q(
                        forum__tracks__user=user, forum__tracks__mark_time__lt=F('last_post_on'),
                        tracks__isnull=True,
                    ),
                )
            )
            if unread_topics.exists():
                break

            # The topics that are marked as read inside the forum for the given user
            # wil be deleted while the forum track associated with the user must be
            # created or updated.
            TopicReadTrack.objects.filter(topic__forum=forum, user=user).delete()
            forum_track, _ = ForumReadTrack.objects.get_or_create(forum=forum, user=user)
            forum_track.save()



from django import template

from machina.core.db.models import get_model
from machina.core.loading import get_class


TopicPollVote = get_model('forum_polls', 'TopicPollVote')

get_anonymous_user_forum_key = get_class(
    'forum_permission.shortcuts', 'get_anonymous_user_forum_key',
)

register = template.Library()


@register.filter
def has_been_completed_by(poll, user):
    """ This will return a boolean indicating if the passed user has already voted in the given
        poll.

    Usage::
        {% if poll|has_been_completed_by:user %}...{% endif %}

    """
    user_votes = TopicPollVote.objects.filter(
        poll_option__poll=poll)
    if user.is_anonymous:
        forum_key = get_anonymous_user_forum_key(user)
        user_votes = user_votes.filter(anonymous_key=forum_key) if forum_key \
            else user_votes.none()
    else:
        user_votes = user_votes.filter(voter=user)
    return user_votes.exists()



import factory
from django.contrib.auth.models import Group, User
from faker import Faker


faker = Faker()


class UserFactory(factory.DjangoModelFactory):
    username = factory.Sequence(lambda n: '{}-{}'.format(faker.user_name(), n))
    email = factory.Sequence(lambda n: 'test{0}@example.com'.format(n))
    password = '1234'
    is_active = True

    class Meta:
        model = User

    @classmethod
    def _prepare(cls, create, **kwargs):
        password = kwargs.pop('password', None)
        user = super()._prepare(create, **kwargs)
        if password:
            user.set_password(password)
            if create:
                user.save()
        return user


class GroupFactory(factory.DjangoModelFactory):
    name = factory.Sequence(lambda n: '{}-{}'.format(str(n), faker.job()))

    class Meta:
        model = Group



"""
    Forum permission checker
    ========================

    This module defines a ``ForumPermissionChecker`` abstraction that allows to check forum
    permissions on specific forum instances.

"""

from django.contrib.auth import get_user_model
from django.db.models import Q

from machina.conf import settings as machina_settings
from machina.core.db.models import get_model


ForumPermission = get_model('forum_permission', 'ForumPermission')
GroupForumPermission = get_model('forum_permission', 'GroupForumPermission')
UserForumPermission = get_model('forum_permission', 'UserForumPermission')


class ForumPermissionChecker:
    """ The ForumPermissionChecker allows to check forum permissions on Forum instances. """

    def __init__(self, user):
        self.user = user
        self._forum_perms_cache = {}

    def has_perm(self, perm, forum):
        """ Checks if the considered user has given permission for the passed forum. """
        if not self.user.is_anonymous and not self.user.is_active:
            # An inactive user cannot have permissions
            return False
        elif self.user and self.user.is_superuser:
            # The superuser have all permissions
            return True
        return perm in self.get_perms(forum)

    def get_perms(self, forum):
        """ Returns the list of permission codenames of all permissions for the given forum. """
        # An inactive user has no permissions.
        if not self.user.is_anonymous and not self.user.is_active:
            return []

        user_model = get_user_model()
        user_groups_related_name = user_model.groups.field.related_query_name()

        if forum.id not in self._forum_perms_cache:
            if self.user and self.user.is_superuser:
                # The superuser has all the permissions.
                perms = list(ForumPermission.objects.values_list('codename', flat=True))
            elif self.user:
                default_auth_forum_perms = (
                    machina_settings.DEFAULT_AUTHENTICATED_USER_FORUM_PERMISSIONS
                )

                user_kwargs_filter = (
                    {'anonymous_user': True} if self.user.is_anonymous
                    else {'user': self.user}
                )

                # Fetches the permissions of the considered user for the given forum.
                user_perms = (
                    UserForumPermission.objects.select_related()
                    .filter(**user_kwargs_filter)
                    .filter(Q(forum__isnull=True) | Q(forum=forum))
                )

                # Computes the list of permissions that are granted for all the forums.
                globally_granted_user_perms = list(
                    filter(lambda p: p.has_perm and p.forum_id is None, user_perms)
                )
                globally_granted_user_perms = [
                    p.permission.codename for p in globally_granted_user_perms
                ]

                # Computes the list of permissions that are granted on a per-forum basis.
                per_forum_granted_user_perms = list(
                    filter(lambda p: p.has_perm and p.forum_id is not None, user_perms)
                )
                per_forum_granted_user_perms = [
                    p.permission.codename for p in per_forum_granted_user_perms
                ]

                # Computes the list of permissions that are not granted on a per-forum basis.
                per_forum_nongranted_user_perms = list(
                    filter(lambda p: not p.has_perm and p.forum_id is not None, user_perms)
                )
                per_forum_nongranted_user_perms = [
                    p.permission.codename for p in per_forum_nongranted_user_perms
                ]

                # If the considered user have no global permissions, the permissions defined by the
                # DEFAULT_AUTHENTICATED_USER_FORUM_PERMISSIONS settings are used instead.
                if self.user.is_authenticated and not globally_granted_user_perms:
                    globally_granted_user_perms = default_auth_forum_perms

                # Finally computes the list of permission codenames that are granted to the user for
                # the considered forum.
                granted_user_perms = [
                    c for c in globally_granted_user_perms if
                    c not in per_forum_nongranted_user_perms
                ]
                granted_user_perms += per_forum_granted_user_perms
                granted_user_perms = set(granted_user_perms)

                perms = granted_user_perms

                # If the user is a registered user, we have to check the permissions of its groups
                # in order to determine the additional permissions they could have.
                if not self.user.is_anonymous:
                    group_perms = (
                        GroupForumPermission.objects.select_related()
                        .filter(**{'group__{}'.format(user_groups_related_name): self.user})
                        .filter(Q(forum__isnull=True) | Q(forum=forum))
                    )

                    globally_granted_group_perms = list(
                        filter(lambda p: p.has_perm and p.forum_id is None, group_perms)
                    )
                    globally_granted_group_perms = [
                        p.permission.codename for p in globally_granted_group_perms
                    ]

                    per_forum_granted_group_perms = list(
                        filter(lambda p: p.has_perm and p.forum_id is not None, group_perms)
                    )
                    per_forum_granted_group_perms = [
                        p.permission.codename for p in per_forum_granted_group_perms
                    ]

                    per_forum_nongranted_group_perms = list(
                        filter(lambda p: not p.has_perm and p.forum_id is not None, group_perms)
                    )
                    per_forum_nongranted_group_perms = [
                        p.permission.codename for p in per_forum_nongranted_group_perms
                    ]

                    granted_group_perms = [
                        c for c in globally_granted_group_perms if
                        c not in per_forum_nongranted_group_perms
                    ]
                    granted_group_perms += per_forum_granted_group_perms
                    granted_group_perms = filter(
                        lambda x: x not in per_forum_nongranted_user_perms, granted_group_perms
                    )
                    granted_group_perms = set(granted_group_perms)

                    # Includes the permissions granted for the user' groups in the initial set of
                    # permission codenames.
                    perms |= granted_group_perms

            self._forum_perms_cache[forum.id] = perms

        return self._forum_perms_cache[forum.id]



"""
    Forum tracking model admin definitions
    ======================================

    This module defines admin classes used to populate the Django administration dashboard.

"""

from django.contrib import admin

from machina.core.db.models import get_model


ForumReadTrack = get_model('forum_tracking', 'ForumReadTrack')
TopicReadTrack = get_model('forum_tracking', 'TopicReadTrack')


class ForumReadTrackAdmin(admin.ModelAdmin):
    """ The Forum Read Track model admin. """

    list_display = ('__str__', 'user', 'forum', 'mark_time',)
    list_filter = ('mark_time',)


class TopicReadTrackAdmin(admin.ModelAdmin):
    """ The Topic Read Track model admin. """

    list_display = ('__str__', 'user', 'topic', 'mark_time',)
    list_filter = ('mark_time',)


admin.site.register(ForumReadTrack, ForumReadTrackAdmin)
admin.site.register(TopicReadTrack, TopicReadTrackAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-04-21 02:31
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('forum_permission', '0002_auto_20160607_0500'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='forumpermission',
            name='name',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-25 19:26
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forum_conversation', '0008_auto_20160903_0512'),
    ]

    operations = [
        migrations.AlterField(
            model_name='topic',
            name='slug',
            field=models.SlugField(max_length=255, verbose_name='Slug'),
        ),
    ]



"""
    Forum conversation forms
    ========================

    This module defines forms provided by the ``forum_conversation`` application.

"""

from django import forms
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import F
from django.utils.translation import ugettext_lazy as _

from machina.conf import settings as machina_settings
from machina.core.db.models import get_model
from machina.core.loading import get_class


Post = get_model('forum_conversation', 'Post')
Topic = get_model('forum_conversation', 'Topic')
TopicPoll = get_model('forum_polls', 'TopicPoll')

PermissionHandler = get_class('forum_permission.handler', 'PermissionHandler')

get_anonymous_user_forum_key = get_class(
    'forum_permission.shortcuts', 'get_anonymous_user_forum_key',
)


class PostForm(forms.ModelForm):
    """ Allows to create or update forum posts. """

    class Meta:
        model = Post
        fields = ['subject', 'content', 'username', 'update_reason', 'enable_signature', ]

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop('user', None)
        self.forum = kwargs.pop('forum', None)
        self.topic = kwargs.pop('topic', None)

        self.perm_handler = PermissionHandler()

        super().__init__(*args, **kwargs)

        # Updates the 'subject' and 'content' fields attributes
        self.fields['subject'].widget.attrs['placeholder'] = _('Enter your subject')
        self.fields['content'].label = _('Message')
        self.fields['content'].widget.attrs['placeholder'] = _('Enter your message')

        # Handles anonymous users
        if self.user and self.user.is_anonymous:
            self.fields['username'].required = True
        else:
            # The 'username' field is not really usefull if the user is authenticated
            del self.fields['username']

        # Handles the definition of a default subject if we are considering an answer
        if not self.instance.pk and self.topic:
            self.fields['subject'].initial = '{} {}'.format(
                machina_settings.TOPIC_ANSWER_SUBJECT_PREFIX,
                self.topic.subject)

        # Delete the 'update_reason' field if we are considering a post update
        if not self.instance.pk:
            del self.fields['update_reason']

        # Inserts a field to allow the user to lock the current topic if he has the permissions to
        # do so.
        if (
            (self.instance.pk or self.topic) and
            self.perm_handler.can_lock_topics(self.forum, self.user)
        ):
            self.fields['lock_topic'] = forms.BooleanField(
                label=_('Lock topic'), required=False,
                initial=self.topic.status == Topic.TOPIC_LOCKED,
            )

    def clean(self):
        """ Validates the form. """
        if not self.instance.pk:
            # Only set user on post creation
            if not self.user.is_anonymous:
                self.instance.poster = self.user
            else:
                self.instance.anonymous_key = get_anonymous_user_forum_key(self.user)
        return super().clean()

    def save(self, commit=True):
        """ Saves the instance. """
        if self.instance.pk:
            # First handle updates
            post = super().save(commit=False)
            post.updated_by = self.user
            post.updates_count = F('updates_count') + 1
        else:
            post = Post(
                topic=self.topic,
                subject=self.cleaned_data['subject'],
                approved=self.perm_handler.can_post_without_approval(self.forum, self.user),
                content=self.cleaned_data['content'],
                enable_signature=self.cleaned_data['enable_signature'])
            if not self.user.is_anonymous:
                post.poster = self.user
            else:
                post.username = self.cleaned_data['username']
                post.anonymous_key = get_anonymous_user_forum_key(self.user)

        # Locks the topic if appropriate.
        lock_topic = self.cleaned_data.get('lock_topic', False)
        if lock_topic:
            self.topic.status = Topic.TOPIC_LOCKED
            self.topic.save()

        if commit:
            post.save()

        return post


class TopicForm(PostForm):
    """ Allows to create or update forum topics. """

    topic_type = forms.ChoiceField(
        label=_('Post topic as'), choices=Topic.TYPE_CHOICES, required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Perform some checks before doing anything
        self.can_add_stickies = self.perm_handler.can_add_stickies(self.forum, self.user)
        self.can_add_announcements = self.perm_handler.can_add_announcements(self.forum, self.user)
        self.can_create_polls = self.perm_handler.can_create_polls(self.forum, self.user)

        if not self.can_add_stickies:
            choices = filter(
                lambda t: t[0] != Topic.TOPIC_STICKY, self.fields['topic_type'].choices,
            )
            self.fields['topic_type'].choices = choices
        if not self.can_add_announcements:
            choices = filter(
                lambda t: t[0] != Topic.TOPIC_ANNOUNCE, self.fields['topic_type'].choices,
            )
            self.fields['topic_type'].choices = choices

        # Append polls fields to the form if the user is allowed to create such things
        if self.can_create_polls:
            self.fields['poll_question'] = forms.CharField(
                label=_('Poll question'), required=False,
                help_text=_(
                    'Enter a question to associate a poll with the topic or leave blank to not '
                    'create a poll.'
                ),
                max_length=TopicPoll._meta.get_field('question').max_length,
            )
            self.fields['poll_max_options'] = forms.IntegerField(
                label=_('Maximum number of poll options per user'), required=False,
                help_text=_('This is the number of options each user may select when voting.'),
                validators=TopicPoll._meta.get_field('max_options').validators,
                initial=TopicPoll._meta.get_field('max_options').default,
            )
            self.fields['poll_duration'] = forms.IntegerField(
                label=_('For how many days the poll should be run?'), required=False,
                help_text=_('Enter 0 or leave blank for a never ending poll.'),
                min_value=0, initial=0,
            )
            self.fields['poll_user_changes'] = forms.BooleanField(
                label=_('Allow re-voting?'), required=False,
                help_text=_('If enabled users are able to change their vote.'),
                initial=False,
            )

        # Set the initial values
        try:
            if hasattr(self.instance, 'topic'):
                self.fields['topic_type'].initial = self.instance.topic.type

                if self.can_create_polls and self.instance.topic.poll is not None:
                    self.fields['poll_question'].initial = self.instance.topic.poll.question
                    self.fields['poll_max_options'].initial = self.instance.topic.poll.max_options
                    self.fields['poll_duration'].initial = self.instance.topic.poll.duration
                    self.fields['poll_user_changes'].initial = self.instance.topic.poll.user_changes
        except ObjectDoesNotExist:
            pass

    def clean(self):
        """ Validates the form. """
        if self.cleaned_data.get('poll_question', None) \
                and not self.cleaned_data.get('poll_max_options', None):
            self.add_error(
                'poll_max_options',
                _('You must set the maximum number of poll options per user when creating polls'),
            )

        super().clean()

    def save(self, commit=True):
        """ Saves the instance. """
        if not self.instance.pk:
            # First, handle topic creation
            if 'topic_type' in self.cleaned_data and len(self.cleaned_data['topic_type']):
                topic_type = self.cleaned_data['topic_type']
            else:
                topic_type = Topic.TOPIC_POST

            topic = Topic(
                forum=self.forum,
                subject=self.cleaned_data['subject'],  # The topic's name is the post's name
                type=topic_type,
                status=Topic.TOPIC_UNLOCKED,
                approved=self.perm_handler.can_post_without_approval(self.forum, self.user),
            )
            if not self.user.is_anonymous:
                topic.poster = self.user
            self.topic = topic
            if commit:
                topic.save()
        else:
            if 'topic_type' in self.cleaned_data and len(self.cleaned_data['topic_type']):
                if self.instance.topic.type != self.cleaned_data['topic_type']:
                    self.instance.topic.type = self.cleaned_data['topic_type']
                    self.instance.topic._simple_save()

        return super().save(commit)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2017-01-20 01:24
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('forum_conversation', '0009_auto_20160925_2126'),
    ]

    operations = [
        migrations.AlterField(
            model_name='topic',
            name='first_post',
            field=models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='forum_conversation.Post', verbose_name='First post'),
        ),
    ]



# -*- coding: utf-8 -*-

import codecs
from os.path import abspath
from os.path import dirname
from os.path import join
from setuptools import find_packages
from setuptools import setup

import machina


def read_relative_file(filename):
    """ Returns contents of the given file, whose path is supposed relative to this module. """
    with codecs.open(join(dirname(abspath(__file__)), filename), encoding='utf-8') as f:
        return f.read()


setup(
    name='django-machina',
    version=machina.__version__,
    author='Morgan Aubert',
    author_email='me@morganaubert.name',
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/ellmetha/django-machina',
    license='BSD',
    description='A Django forum engine for building powerful community driven websites.',
    long_description=read_relative_file('README.rst'),
    keywords='django forum board messages',
    zip_safe=False,
    install_requires=[
        'django>=2.0',

        # Django-mptt is required to handle the tree hierarchy of nested forums
        'django-mptt>=0.8',

        # Machina uses Django-haystack to provide search support
        'django-haystack>=2.1',

        # Pillow is required for image fields
        'pillow>=2.2',

        # Machina uses Markdown by default as a syntax for forum messages ; but you can change this
        'markdown2>=2.3',

        # Machina's default templates use django-widget-tweaks to render form fields ; but you can
        # override this
        'django-widget-tweaks>=1.4',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Message Boards',
    ],
)



"""
    Forum tracking signal receivers
    ===============================

    This module defines signal receivers.

"""

from django.dispatch import receiver

from machina.core.loading import get_class


topic_viewed = get_class('forum_conversation.signals', 'topic_viewed')


@receiver(topic_viewed)
def update_user_trackers(sender, topic, user, request, response, **kwargs):
    """ Receiver to mark a topic being viewed as read.

    This can result in marking the related forum tracker as read.

    """
    TrackingHandler = get_class('forum_tracking.handler', 'TrackingHandler')  # noqa
    track_handler = TrackingHandler()
    track_handler.mark_topic_read(topic, user)



"""
    Forum member views
    ==================

    This module defines views provided by the ``forum_member`` application.

"""

from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from django.views.generic import DetailView, ListView, UpdateView
from django.views.generic.detail import BaseDetailView, SingleObjectTemplateResponseMixin

from machina.conf import settings as machina_settings
from machina.core.db.models import get_model
from machina.core.loading import get_class


Forum = get_model('forum', 'Forum')
ForumProfile = get_model('forum_member', 'ForumProfile')
Post = get_model('forum_conversation', 'Post')
Topic = get_model('forum_conversation', 'Topic')

ForumProfileForm = get_class('forum_member.forms', 'ForumProfileForm')

PermissionRequiredMixin = get_class('forum_permission.viewmixins', 'PermissionRequiredMixin')


class UserPostsView(ListView):
    """ Provides a list of all the posts submitted by a given a user. """

    context_object_name = 'posts'
    paginate_by = machina_settings.PROFILE_POSTS_NUMBER_PER_PAGE
    template_name = 'forum_member/user_posts_list.html'
    user_pk_url_kwarg = 'pk'

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        context['poster'] = self.poster
        return context

    def get_queryset(self):
        """ Returns the list of items for this view. """
        # Determines the forums that can be accessed by the current user
        forums = self.request.forum_permission_handler.get_readable_forums(
            Forum.objects.all(), self.request.user,
        )

        # Returns the posts submitted by the considered user.
        return (
            Post.objects
            .select_related('topic', 'topic__forum', 'poster')
            .prefetch_related('poster__forum_profile')
            .filter(poster=self.poster, topic__forum__in=forums)
            .order_by('-created')
        )

    @cached_property
    def poster(self):
        """ Returns the considered user. """
        user_model = get_user_model()
        return get_object_or_404(user_model, pk=self.kwargs[self.user_pk_url_kwarg])


class ForumProfileDetailView(DetailView):
    """ Shows a user's forum profile. """

    context_object_name = 'profile'
    template_name = 'forum_member/forum_profile_detail.html'

    def get_queryset(self):
        """ Returns the list of items for this view. """
        user_model = get_user_model()
        return user_model.objects.all()

    def get_object(self, queryset=None):
        """ Returns the considered object. """
        user = super().get_object(queryset)
        profile, dummy = ForumProfile.objects.get_or_create(user=user)
        return profile

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)

        # Computes the number of topics added by the considered member
        context['topics_count'] = (
            Topic.objects.filter(approved=True, poster=self.object.user).count()
        )

        # Fetches the recent posts added by the considered user
        forums = self.request.forum_permission_handler.get_readable_forums(
            Forum.objects.all(), self.request.user,
        )
        recent_posts = (
            Post.approved_objects
            .select_related('topic', 'topic__forum')
            .filter(topic__forum__in=forums, poster=self.object.user)
            .order_by('-created')
        )
        context['recent_posts'] = recent_posts[:machina_settings.PROFILE_RECENT_POSTS_NUMBER]

        return context


class ForumProfileUpdateView(LoginRequiredMixin, UpdateView):
    """ Allows the current user to update its forum profile. """

    form_class = ForumProfileForm
    template_name = 'forum_member/forum_profile_update.html'
    success_message = _('The profile has been edited successfully.')

    def get_object(self, queryset=None):
        """ Returns the considered object. """
        profile, dummy = ForumProfile.objects.get_or_create(user=self.request.user)
        return profile

    def form_valid(self, form):
        """ Handles a valid form. """
        response = super().form_valid(form)
        messages.success(self.request, self.success_message)
        return response

    def get_success_url(self):
        """ Returns the success URL to redirect the user to. """
        return reverse('forum_member:profile_update')


class TopicSubscribeView(
    LoginRequiredMixin, PermissionRequiredMixin, SingleObjectTemplateResponseMixin, BaseDetailView,
):
    """ Allows a user to subscribe to a specific topic. """

    model = Topic
    success_message = _('You subscribed to this topic successfully.')
    template_name = 'forum_member/topic_subscribe.html'

    def subscribe(self, request, *args, **kwargs):
        """ Performs the subscribe action. """
        self.object = self.get_object()
        self.object.subscribers.add(request.user)
        messages.success(self.request, self.success_message)
        return HttpResponseRedirect(self.get_success_url())

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        return self.subscribe(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        context['topic'] = self.object
        context['forum'] = self.object.forum
        return context

    def get_success_url(self):
        """ Returns the success URL to redirect the user to. """
        return reverse(
            'forum_conversation:topic',
            kwargs={
                'forum_slug': self.object.forum.slug,
                'forum_pk': self.object.forum.pk,
                'slug': self.object.slug,
                'pk': self.object.pk,
            },
        )

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permission check. """
        return self.request.forum_permission_handler.can_subscribe_to_topic(obj, user)


class TopicUnsubscribeView(
    LoginRequiredMixin, PermissionRequiredMixin, SingleObjectTemplateResponseMixin, BaseDetailView,
):
    """
    Allows a user to unsubscribe from a specific topic.
    """

    model = Topic
    success_message = _('You unsubscribed from this topic successfully.')
    template_name = 'forum_member/topic_unsubscribe.html'

    def unsubscribe(self, request, *args, **kwargs):
        """ Performs the unsubscribe action. """
        self.object = self.get_object()
        self.object.subscribers.remove(request.user)
        messages.success(self.request, self.success_message)
        return HttpResponseRedirect(self.get_success_url())

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        return self.unsubscribe(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        context['topic'] = self.object
        context['forum'] = self.object.forum
        return context

    def get_success_url(self):
        """ Returns the success URL to redirect the user to. """
        return reverse(
            'forum_conversation:topic',
            kwargs={
                'forum_slug': self.object.forum.slug,
                'forum_pk': self.object.forum.pk,
                'slug': self.object.slug,
                'pk': self.object.pk,
            },
        )

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permission check. """
        return self.request.forum_permission_handler.can_unsubscribe_from_topic(obj, user)


class TopicSubscriptionListView(LoginRequiredMixin, ListView):
    """ Provides a list of all topics to which the current user has subscribed. """

    context_object_name = 'topics'
    model = Topic
    paginate_by = machina_settings.FORUM_TOPICS_NUMBER_PER_PAGE
    template_name = 'forum_member/subscription_topic_list.html'

    def get_queryset(self):
        """ Returns the list of items for this view. """
        return (
            self.request.user.topic_subscriptions
            .select_related('forum', 'poster', 'last_post', 'last_post__poster')
            .all()
        )



"""
    Forum conversation abstract models
    ==================================

    This module defines abstract models provided by the ``forum_conversation`` application.

"""

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.utils.encoding import force_text
from django.utils.text import slugify
from django.utils.translation import ugettext_lazy as _

from machina.conf import settings as machina_settings
from machina.core import validators
from machina.core.loading import get_class
from machina.models.abstract_models import DatedModel
from machina.models.fields import MarkupTextField


ApprovedManager = get_class('forum_conversation.managers', 'ApprovedManager')


class AbstractTopic(DatedModel):
    """ Represents a forum topic. """

    forum = models.ForeignKey(
        'forum.Forum', related_name='topics', on_delete=models.CASCADE,
        verbose_name=_('Topic forum'),
    )
    poster = models.ForeignKey(
        settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.CASCADE,
        verbose_name=_('Poster'),
    )

    # The subject of the thread should correspond to the one associated with the first post.
    subject = models.CharField(max_length=255, verbose_name=_('Subject'))
    slug = models.SlugField(max_length=255, verbose_name=_('Slug'))

    # Sticky, Announce, Global topic or Default topic ; that's what a topic can be.
    TOPIC_POST, TOPIC_STICKY, TOPIC_ANNOUNCE = 0, 1, 2
    TYPE_CHOICES = (
        (TOPIC_POST, _('Default topic')),
        (TOPIC_STICKY, _('Sticky')),
        (TOPIC_ANNOUNCE, _('Announce')),
    )
    type = models.PositiveSmallIntegerField(
        choices=TYPE_CHOICES, db_index=True, verbose_name=_('Topic type'),
    )

    # A topic can be locked, unlocked or moved.
    TOPIC_UNLOCKED, TOPIC_LOCKED, TOPIC_MOVED = 0, 1, 2
    STATUS_CHOICES = (
        (TOPIC_UNLOCKED, _('Topic unlocked')),
        (TOPIC_LOCKED, _('Topic locked')),
        (TOPIC_MOVED, _('Topic moved')),
    )
    status = models.PositiveIntegerField(
        choices=STATUS_CHOICES, db_index=True, verbose_name=_('Topic status'),
    )

    # A topic can be approved before publishing ; defaults to True. The value of this flag
    # should correspond to the one associated with the first post.
    approved = models.BooleanField(default=True, db_index=True, verbose_name=_('Approved'))

    # The number of posts included in this topic (only those that are approved).
    posts_count = models.PositiveIntegerField(
        editable=False, blank=True, default=0, verbose_name=_('Posts count'),
    )

    # The number of time the topic has been viewed.
    views_count = models.PositiveIntegerField(
        editable=False, blank=True, default=0, verbose_name=_('Views count'),
    )

    # The date of the latest post.
    last_post_on = models.DateTimeField(blank=True, null=True, verbose_name=_('Last post added on'))

    # The first post and the last post of the topic. The first post can be unnaproved. The last post
    # must be approved.
    first_post = models.ForeignKey(
        'forum_conversation.Post', editable=False, related_name='+', blank=True, null=True,
        on_delete=models.SET_NULL, verbose_name=_('First post'),
    )
    last_post = models.ForeignKey(
        'forum_conversation.Post', editable=False, related_name='+', blank=True, null=True,
        on_delete=models.SET_NULL, verbose_name=_('Last post'),
    )

    # Many users can subscribe to this topic
    subscribers = models.ManyToManyField(
        settings.AUTH_USER_MODEL, related_name='topic_subscriptions', blank=True,
        verbose_name=_('Subscribers'),
    )

    objects = models.Manager()
    approved_objects = ApprovedManager()

    class Meta:
        abstract = True
        app_label = 'forum_conversation'
        ordering = ['-type', '-last_post_on', ]
        get_latest_by = 'last_post_on'
        verbose_name = _('Topic')
        verbose_name_plural = _('Topics')

    def __str__(self):
        return self.first_post.subject if self.first_post is not None else self.subject

    @property
    def is_topic(self):
        """ Returns ``True`` if the topic is a default topic. """
        return self.type == self.TOPIC_POST

    @property
    def is_sticky(self):
        """ Returns ``True`` if the topic is a sticky topic. """
        return self.type == self.TOPIC_STICKY

    @property
    def is_announce(self):
        """ Returns ``True`` if the topic is an announce. """
        return self.type == self.TOPIC_ANNOUNCE

    @property
    def is_locked(self):
        """ Returns ``True`` if the topic is locked. """
        return self.status == self.TOPIC_LOCKED

    def has_subscriber(self, user):
        """ Returns ``True`` if the given user is a subscriber of this topic. """
        if not hasattr(self, '_subscribers'):
            self._subscribers = list(self.subscribers.all())
        return user in self._subscribers

    def clean(self):
        """ Validates the topic instance. """
        super().clean()
        if self.forum.is_category or self.forum.is_link:
            raise ValidationError(
                _('A topic can not be associated with a category or a link forum')
            )

    def save(self, *args, **kwargs):
        """ Saves the topic instance. """
        # It is vital to track the changes of the forum associated with a topic in order to
        # maintain counters up-to-date.
        old_instance = None
        if self.pk:
            old_instance = self.__class__._default_manager.get(pk=self.pk)

        # Update the slug field
        self.slug = slugify(force_text(self.subject), allow_unicode=True)

        # Do the save
        super().save(*args, **kwargs)

        # If any change has been made to the parent forum, trigger the update of the counters
        if old_instance and old_instance.forum != self.forum:
            self.update_trackers()
            # The previous parent forum counters should also be updated
            if old_instance.forum:
                old_forum = old_instance.forum
                old_forum.refresh_from_db()
                old_forum.update_trackers()

    def _simple_save(self, *args, **kwargs):
        """ Simple wrapper around the standard save method.

        Calls the parent save method in order to avoid the checks for topic forum changes which can
        result in triggering a new update of the counters associated with the current topic.

        This allow the database to not be hit by such checks during very common and regular
        operations such as those provided by the update_trackers function; indeed these operations
        will never result in an update of a topic's forum.

        """
        super().save(*args, **kwargs)

    def delete(self, using=None):
        """ Deletes the forum instance. """
        super().delete(using)
        self.forum.update_trackers()

    def update_trackers(self):
        """ Updates the denormalized trackers associated with the topic instance. """
        self.posts_count = self.posts.filter(approved=True).count()
        first_post = self.posts.all().order_by('created').first()
        last_post = self.posts.filter(approved=True).order_by('-created').first()
        self.first_post = first_post
        self.last_post = last_post
        self.last_post_on = last_post.created if last_post else None
        self._simple_save()
        # Trigger the forum-level trackers update
        self.forum.update_trackers()


class AbstractPost(DatedModel):
    """ Represents a forum post. A forum post is always linked to a topic. """

    topic = models.ForeignKey(
        'forum_conversation.Topic', related_name='posts', on_delete=models.CASCADE,
        verbose_name=_('Topic'),
    )
    poster = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name='posts',
        blank=True, null=True, on_delete=models.CASCADE, verbose_name=_('Poster'),
    )
    anonymous_key = models.CharField(
        max_length=100, blank=True, null=True, verbose_name=_('Anonymous user forum key'),
    )

    # Each post can have its own subject. The subject of the thread corresponds to the
    # one associated with the first post
    subject = models.CharField(verbose_name=_('Subject'), max_length=255)

    # Content
    content = MarkupTextField(
        validators=[
            validators.NullableMaxLengthValidator(machina_settings.POST_CONTENT_MAX_LENGTH),
        ],
        verbose_name=_('Content'),
    )

    # Username: if the user creating a topic post is not authenticated, he must enter a username
    username = models.CharField(max_length=155, blank=True, null=True, verbose_name=_('Username'))

    # A post can be approved before publishing ; defaults to True
    approved = models.BooleanField(default=True, db_index=True, verbose_name=_('Approved'))

    # The user can choose if they want to display their signature with the content of the post
    enable_signature = models.BooleanField(
        default=True, db_index=True, verbose_name=_('Attach a signature'),
    )

    # A post can be edited for several reason (eg. moderation) ; the reason why it has been
    # updated can be specified
    update_reason = models.CharField(
        max_length=255, blank=True, null=True, verbose_name=_('Update reason'),
    )

    # Tracking data
    updated_by = models.ForeignKey(
        settings.AUTH_USER_MODEL, editable=False, blank=True, null=True, on_delete=models.SET_NULL,
        verbose_name=_('Lastly updated by'),
    )
    updates_count = models.PositiveIntegerField(
        editable=False, blank=True, default=0, verbose_name=_('Updates count'),
    )

    objects = models.Manager()
    approved_objects = ApprovedManager()

    class Meta:
        abstract = True
        app_label = 'forum_conversation'
        ordering = ['created', ]
        get_latest_by = 'created'
        verbose_name = _('Post')
        verbose_name_plural = _('Posts')

    def __str__(self):
        return self.subject

    @property
    def is_topic_head(self):
        """ Returns ``True`` if the post is the first post of the topic. """
        return self.topic.first_post.id == self.id if self.topic.first_post else False

    @property
    def is_topic_tail(self):
        """ Returns ``True`` if the post is the last post of the topic. """
        return self.topic.last_post.id == self.id if self.topic.last_post else False

    @property
    def is_alone(self):
        """ Returns ``True`` if the post is the only single post of the topic. """
        return self.topic.posts.count() == 1

    @property
    def position(self):
        """ Returns an integer corresponding to the position of the post in the topic. """
        position = self.topic.posts.filter(Q(created__lt=self.created) | Q(id=self.id)).count()
        return position

    def clean(self):
        """ Validates the post instance. """
        super().clean()

        # At least a poster (user) or a session key must be associated with
        # the post.
        if self.poster is None and self.anonymous_key is None:
            raise ValidationError(
                _('A user id or an anonymous key must be associated with a post.'),
            )
        if self.poster and self.anonymous_key:
            raise ValidationError(
                _('A user id or an anonymous key must be associated with a post, but not both.'),
            )

        if self.anonymous_key and not self.username:
            raise ValidationError(_('A username must be specified if the poster is anonymous'))

    def save(self, *args, **kwargs):
        """ Saves the post instance. """
        new_post = self.pk is None
        super().save(*args, **kwargs)

        # Ensures that the subject of the thread corresponds to the one associated
        # with the first post. Do the same with the 'approved' flag.
        if (new_post and self.topic.first_post is None) or self.is_topic_head:
            if self.subject != self.topic.subject or self.approved != self.topic.approved:
                self.topic.subject = self.subject
                self.topic.approved = self.approved

        # Trigger the topic-level trackers update
        self.topic.update_trackers()

    def delete(self, using=None):
        """ Deletes the post instance. """
        if self.is_alone:
            # The default way of operating is to trigger the deletion of the associated topic
            # only if the considered post is the only post embedded in the topic
            self.topic.delete()
        else:
            super(AbstractPost, self).delete(using)
            self.topic.update_trackers()



import inspect

from django import template

from machina.core.loading import get_class


PermissionHandler = get_class('forum_permission.handler', 'PermissionHandler')

register = template.Library()


@register.simple_tag(takes_context=True)
def get_permission(context, method, *args, **kwargs):
    """ This will return a boolean indicating if the considered permission is granted for the passed
        user.

    Usage::

        {% get_permission 'can_access_moderation_panel' request.user as var %}

    """
    request = context.get('request', None)
    perm_handler = request.forum_permission_handler if request else PermissionHandler()

    allowed_methods = inspect.getmembers(perm_handler, predicate=inspect.ismethod)
    allowed_method_names = [a[0] for a in allowed_methods if not a[0].startswith('_')]

    if method not in allowed_method_names:
        raise template.TemplateSyntaxError(
            'Only the following methods are allowed through '
            'this templatetag: {}'.format(allowed_method_names))

    perm_method = getattr(perm_handler, method)
    return perm_method(*args, **kwargs)



from django.core import validators


class NullableMaxLengthValidator(validators.MaxLengthValidator):
    """ Provides a way to not validate an input if the max length is None """

    def __call__(self, value):
        if self.limit_value is None:
            # If the limit value is None, this means that there is no
            # limit value at all. The default validation process is not
            # performed.
            return
        super().__call__(value)



import factory
from faker import Faker

from machina.core.db.models import get_model
from machina.test.factories.conversation import PostFactory


faker = Faker()

Attachment = get_model('forum_attachments', 'Attachment')


class AttachmentFactory(factory.DjangoModelFactory):
    post = factory.SubFactory(PostFactory)
    comment = faker.text(max_nb_chars=255)

    class Meta:
        model = Attachment
